% /*************************************************************************************
%
%    Project Name:  Golay Correlator
%    File Name:     GolayCorrelator.m
%    Authors:       Prem Kiran Nerella
%    Version:       1.0
%    History:       May 2014 created
%
%  *************************************************************************************
%    Description:Operates in two mode       
%    ***********
%    NORMAL MODE   
%    ***********
%    function performs Golay Correlation for differnt code lengths [128, 64, 32]
%    specified in IEEE 802.11ad Std.
%    
%    [GaSeq GbSeq] = GolayCorrelator(InpSeq, CodeLen)
%
%    Inputs
%
%       1. InpSeq   - input data samples
%       2. CodeLen  - specify the code length of correlation sequence,
%                     takes values 128 or 64 or 32%
%   Outputs
%   
%       1. GaSeq  - correlation output by Ga_[CodeLen] sequence
%       2. GbSeq  - correlation output by Gb_[CodeLen] sequence
%   
%   **************
%   STAGGERED MODE
%   **************
%   Staggered input correlation can be performed on larger length
%   input data samples. This can be done by calling the function
%   multiple times in a loop. 
%       
%   Additional PropertyType Inputs
%       
%       1.FirstIteration - (Integer) specifies if the current iterations is the first
%       2.State          - (handle) specifies handle to correlator starting state
%
%   Outputs
%
%       1. State         - handle to correlator current state
%
%   [GaSeq GbSeq State] = GolayCorrelator(...,'FirstIteration',1,...)
%   configure 'FirstIteration' property in staggered correlation mode.
%   specify a value 1 if the current iteration is the first or else a value 0
%
%   [GaSeq GbSeq BufferOut] = GolayCorrelator(...,'State',State)
%   configure 'State' property in staggered correlation mode.
%   specify a handle obtained from previous iteration.
%
%   Example
%       Len     = 1280;
%       Samples = ones(1,Len);
%       CodeLen = 128;
%       iterations = floor(Len/CodeLen); 
%       State = [];Ga = []; Gb = [];
%       for k=0:iterations-1
%           InpSeq = Samples(k*CodeLen+1 : (k+1)*CodeLen);
%           [GaSeq GbSeq StateOut] = GolayCorrelator(InpSeq,CodeLen,'FirstIteration',(k==0),'State',State);
%           State = StateOut;
%           Ga = [Ga GaSeq];Gb = [Gb GbSeq];
%       end
%
%  *************************************************************************************/
function [GaSeq GbSeq varargout] = GolayCorrelator(InpSeq, CodeLen, varargin)

%read varargin
nin = max(nargin,2)-2;
if(nin == 0)
    %Default values
    Initialize    = 1;
elseif( (nin > 0) && (mod(nin,2) == 0) )
    %Read varargin into variables
    CheckStr = {'FirstIteration','State'};
    maxChk = size(CheckStr,2);
    % Read property tags
    Tindx = [];
    for k = 1:2:nin
        for n = 1:maxChk
            if( strcmpi(CheckStr(n),varargin{k} ))
                Tindx = [Tindx n];
            end
        end
    end    
    if( length(Tindx) ~= (nin/2) )
        error('invalid input parameters');
    end
    %Read values associated with tags
    Initialize = 1;Buf = [];
    for k = 1:length(Tindx)
        n = k*2;
        if(Tindx(k) == 1)
            Initialize    = varargin{n};
        elseif(Tindx(k) == 2)
            Buf     = varargin{n};
        end
    end
else
    error('insufficient input parameters');
end

%CodeLen specifies the Goly correlation sequence length
%Correlator Delay Values(Dk) and Weights(Wk) are specified
%Supported CodeLen {'128','64','32'}
if(CodeLen==128)
    %Gx128
    Dk = [1 8 2 4 16 32 64];
    Wk = [-1 -1 -1 -1 +1 -1 -1];
elseif(CodeLen==64)
    %Gx64
    Dk= [2 1 4 8 16 32];
    Wk =[+1 +1 -1 -1 +1 -1];
elseif(CodeLen==32)
    %Gx32
    Dk=[1 4 8 2 16];
    Wk =[-1 +1 -1 +1 -1];
end

%Create buffers when property 'Initialize' is set to value 1 or else
%read the Data Buffers from varargin when property 'Initialize' is set to value other than 1
if(Initialize == 1) B0_rb = ring_buffer([], [], Dk(1), 'init'); else B0_rb = Buf.B0_rb; end
if(Initialize == 1) B1_rb = ring_buffer([], [], Dk(2), 'init'); else B1_rb = Buf.B1_rb; end
if(Initialize == 1) B2_rb = ring_buffer([], [], Dk(3), 'init'); else B2_rb = Buf.B2_rb; end
if(Initialize == 1) B3_rb = ring_buffer([], [], Dk(4), 'init'); else B3_rb = Buf.B3_rb; end
if(Initialize == 1) B4_rb = ring_buffer([], [], Dk(5), 'init'); else B4_rb = Buf.B4_rb; end    
if(CodeLen > 32)
    if(Initialize == 1) B5_rb = ring_buffer([], [], Dk(6), 'init'); else B5_rb = Buf.B5_rb; end
end
if(CodeLen > 64)
    if(Initialize == 1) B6_rb = ring_buffer([], [], Dk(7), 'init'); else B6_rb = Buf.B6_rb; end
end
%Correlator Buffer Flag 'Bx_f' indicate whether Buffer (elements) Count is equal to 'Dk(x)' 
if(Initialize == 1) B0_f = 0; else B0_f = Buf.B0_f; end
if(Initialize == 1) B1_f = 0; else B1_f = Buf.B1_f; end
if(Initialize == 1) B2_f = 0; else B2_f = Buf.B2_f; end
if(Initialize == 1) B3_f = 0; else B3_f = Buf.B3_f; end
if(Initialize == 1) B4_f = 0; else B4_f = Buf.B4_f; end
if(CodeLen > 32)
    if(Initialize == 1) B5_f = 0; else B5_f = Buf.B5_f; end
end
if(CodeLen > 64)
    if(Initialize == 1) B6_f = 0; else B6_f = Buf.B6_f; end
end

%Input samples length
N = length(InpSeq);
%Initialize output
GaSeq = zeros(1,N);
GbSeq = zeros(1,N);

%Golay Correation 
% /*************************************************************************************
%Golay Sequences specified in IEEE 802.11ad Std are defined as described below

%A_[0] = 1;
%B_[0] = 1;

%A_[k](n) = Wk * A_[k-1](n) + B_[k-1](n-Dk);
%B_[k](n) = Wk * A_[k-1](n) - B_[k-1](n-Dk);

%Note that A_[k](n) , B[k](n) are zero for n < 0

%Ga_[128](n) = A_[7](128 - n)
%Gb_[128](n) = B_[7](128 - n)

%Ga_[64](n) = A_[6](64 - n)
%Gb_[64](n) = B_[6](64 - n)

%Ga_[32](n) = A_[5](32 - n)
%Gb_[32](n) = B_[5](32 - n)
%  *************************************************************************************/
for n = 1:N

    m = n-1;
    %A_[0] & B_[0]
    A_0 = InpSeq(n);B0 = A_0;
    if(B0_f == 0) if((m-Dk(1)) >= 0) [B0_rb B_0] = ring_buffer(B0_rb, [], [], 'pop'); B0_f = 1; else B_0 = 0; end; else [B0_rb B_0] = ring_buffer(B0_rb, [], [], 'pop'); end        
    A1 = Wk(1)*A_0 + B_0;
    B1 = Wk(1)*A_0 - B_0;
    %A_[1] & B_[1]
    A_1 = A1;
    if(B1_f == 0) if((m-Dk(2)) >= 0) [B1_rb B_1] = ring_buffer(B1_rb, [], [], 'pop'); B1_f = 1; else B_1 = 0; end; else [B1_rb B_1] = ring_buffer(B1_rb, [], [], 'pop'); end 
    A2 = Wk(2)*A_1 + B_1;
    B2 = Wk(2)*A_1 - B_1;
    %A_[2] & B_[2]
    A_2 = A2;
    if(B2_f == 0) if((m-Dk(3)) >= 0) [B2_rb B_2] = ring_buffer(B2_rb, [], [], 'pop'); B2_f = 1; else B_2 = 0; end; else [B2_rb B_2] = ring_buffer(B2_rb, [], [], 'pop'); end
    A3 = Wk(3)*A_2 + B_2;
    B3 = Wk(3)*A_2 - B_2;
    %A_[3] & B_[3]
    A_3 = A3;
    if(B3_f == 0) if((m-Dk(4)) >= 0) [B3_rb B_3] = ring_buffer(B3_rb, [], [], 'pop'); B3_f = 1; else B_3 = 0; end; else [B3_rb B_3] = ring_buffer(B3_rb, [], [], 'pop'); end
    A4 = Wk(4)*A_3 + B_3;
    B4 = Wk(4)*A_3 - B_3;
    %A_[4] & B_[4]
    A_4 = A4;
    if(B4_f == 0) if((m-Dk(5)) >= 0) [B4_rb B_4] = ring_buffer(B4_rb, [], [], 'pop'); B4_f = 1; else B_4 = 0; end; else [B4_rb B_4] = ring_buffer(B4_rb, [], [], 'pop'); end
    A5 = Wk(5)*A_4 + B_4;
    B5 = Wk(5)*A_4 - B_4;
    %A_[5] & B_[5]
    if(CodeLen > 32)
        A_5 = A5;
        if(B5_f == 0) if((m-Dk(6)) >= 0) [B5_rb B_5] = ring_buffer(B5_rb, [], [], 'pop'); B5_f = 1; else B_5 = 0; end; else [B5_rb B_5] = ring_buffer(B5_rb, [], [], 'pop'); end
        A6 = Wk(6)*A_5 + B_5;
        B6 = Wk(6)*A_5 - B_5;
    end
    %A_[6] & B_[6]
    if(CodeLen > 64)
        A_6 = A6;
        if(B6_f == 0) if((m-Dk(7)) >= 0) [B6_rb B_6] = ring_buffer(B6_rb, [], [], 'pop'); B6_f = 1; else B_6 = 0; end; else [B6_rb B_6] = ring_buffer(B6_rb, [], [], 'pop'); end
        A7 = Wk(7)*A_6 + B_6;
        B7 = Wk(7)*A_6 - B_6;
    end
    %A_[7] & B_[7]
    [B0_rb dataOut] = ring_buffer(B0_rb, B0, [], 'push');
    [B1_rb dataOut] = ring_buffer(B1_rb, B1, [], 'push');
    [B2_rb dataOut] = ring_buffer(B2_rb, B2, [], 'push');
    [B3_rb dataOut] = ring_buffer(B3_rb, B3, [], 'push');
    [B4_rb dataOut] = ring_buffer(B4_rb, B4, [], 'push');
    GaSeq(n) = A5;
    GbSeq(n) = B5;
    if(CodeLen > 32)
        [B5_rb dataOut] = ring_buffer(B5_rb, B5, [], 'push');
        GaSeq(n) = A6;
        GbSeq(n) = B6;
    end
    if(CodeLen > 64)
        [B6_rb dataOut] = ring_buffer(B6_rb, B6, [], 'push');
        GaSeq(n) = A7;
        GbSeq(n) = B7;
    end
end
%Normalize
GaSeq = GaSeq/CodeLen;
GbSeq = GbSeq/CodeLen;

nout = max(nargout,2)-2;
if(nout == 0)
elseif(nout == 1)
    Buf.B0_rb = B0_rb; Buf.B1_rb = B1_rb; Buf.B2_rb = B2_rb; Buf.B3_rb = B3_rb; Buf.B4_rb = B4_rb;
    if(CodeLen > 32) Buf.B5_rb = B5_rb; end
    if(CodeLen > 64) Buf.B6_rb = B6_rb; end
    Buf.B0_f = B0_f; Buf.B1_f = B1_f; Buf.B2_f = B2_f; Buf.B3_f = B3_f; Buf.B4_f = B4_f;
    if(CodeLen > 32) Buf.B5_f = B5_f; end
    if(CodeLen > 64) Buf.B6_f = B6_f; end
    
    varargout{1} = Buf;
else
    error('insufficient output parameters');
end

return
